!function(){
   /**
    * Wrap a function with before, after functions around it.
    */
   function wrap(fn, before, after) {
      var ignoredKeys = { $meta: 1, prototype: 1 };

      function closure() {
         // Invoke before, if it returns false, skip fn() and after()
         if (before.apply(this, arguments) === Infinity) { return; }

         // Invoke fn, save return value
         var ret = fn.apply(this, arguments);

         // Invoke after()
         after.apply(this, arguments);

         // Return ret
         return ret;
      };

      // Wrap fn but keep its properties
      each(fn, function(key, value) {
         if ( !ignoredKeys[key]) {
            closure[key] = value;
         }
      });
      return closure;
   }


   /**
    * Apply Aspect Oriented Programming into a class.
    */
   function pointcuts(clazz, opts) {
      if ( !(isClass(clazz) || isMap(clazz)) || !isMap(opts)) {
         throw "Invalid params";
      }

      var isClass    = isFunction(clazz),
          isInstance = !isClass,
          bindTo     = isClass ? clazz.prototype : clazz;

      each(opts, function(method, pointcuts) {
         // Support sugar syntax for before
         pointcuts  = isFunction(pointcuts) ? { before: pointcuts } : pointcuts;

         var before = (pointcuts.before || noop),
             after  = (pointcuts.after || noop);

         // Checking
         if ( !isFunction(before)) {
            throw "Invalid " + method + ".before()";
         }
         if ( !isFunction(after)) {
            throw "Invalid " + method + ".after()";
         }

         if (isInstance) {
            if (isFunction(bindTo[method])) {
               bindTo[method] = wrap(bindTo[method], before, after);
            } else {
               throw method + " is not a function";
            }
         } else {
            if (isMap(clazz.$meta) && isIdentifier(clazz.$meta.name) && method === clazz.$meta.name) {
               var proto       = clazz.prototype, cls = clazz,     // backup prototype. cls: backup
                   constructor = wrap(clazz, before, after);       // new constructor

               // Restore prototype, no data lost
               if (clazz.$meta.namespace) {
                  clazz = clazz.$meta.namespace[method] = constructor;
               } else {
                  clazz = exports(method, constructor);
               }

               clazz.prototype = proto;
               clazz.prototype.constructor = constructor;

               // Re-bind static properties, except prototype
               each(cls, function(property, value) {
                  if (property !== "prototype") {
                     clazz[property] = value;
                  }
               });
            } else {
               if (isFunction(bindTo[method])) {
                  bindTo[method] = wrap(bindTo[method], before, after);

                  // Also is a static function? Update static pointer
                  if (isFunction(clazz[method])) {
                     clazz[method] = bindTo[method];
                  }
               } else {
                  throw method + " is not a function";
               }
            }
         }
      });
   }

   /**
    * Class plugins. The repository for Class()'s processing plugins.
    */
   Class.plugins = {

      /*
       * Plug-in: pointcuts.
       * Purpose: Support Aspect Oriented Programming in Class().
       */
      pointcuts: function(clazz, opts) {
         if (isMap(opts.$meta.pointcuts)) {
            pointcuts(clazz, opts.$meta.pointcuts);
         }
      }
   };

   if (typeof module !== "undefined" && module.exports) {
      var jsface = require("./jsface");
      jsface.pointcut = overload;
   }

   var extend     = jsface.extend,
       each       = jsface.each,
       isMap      = jsface.isMap,
       isArray    = jsface.isArray,
       isFunction = jsface.isFunction,
       isString   = jsface.isString,
       isClass    = jsface.isClass;

}();
